/************************************************************************
* Tetris-C	(A Console Application) 									*
* Written in C Using Dev C++, this program emulates my own version of 	*
* Tetris using a procedural programming approach to allow the player to	*
* play a game of Tetris utlizing ascii values to form the tetrominoes.	*
* By Chad Jordan														*
* 07/04/2008															*
************************************************************************/

#include <stdio.h> 
#include <stdlib.h> 
#include <time.h> 
#include <conio.h>
#include <windows.h>

#define DISPLAY_WIDTH (BOARD_WIDTH * 2 + 14)
#define DISPLAY_HEIGHT (BOARD_HEIGHT + 4)
#define BOARD_WIDTH 10
#define BOARD_HEIGHT 20
#define PIECE_SIZE 4 

//Add a new variable to hold the next piece and color.
int nextPiece[PIECE_SIZE][PIECE_SIZE]; 
int nextColor; 

//Tetromino shapes, represented as 4x4 matrices.
int tetrominoes[7][PIECE_SIZE][PIECE_SIZE] = {
    {
        {1, 1, 1, 1},
        {0, 0, 0, 0},
        {0, 0, 0, 0},
        {0, 0, 0, 0}
    },
    {
        {1, 1},
        {1, 1}
    },
    {
        {0, 1, 0},
        {1, 1, 1},
        {0, 0, 0}
    },
    {
        {1, 1, 0},
        {0, 1, 1},
        {0, 0, 0}
    },
    {
        {0, 1, 1},
        {1, 1, 0},
        {0, 0, 0}
    },
    {
        {1, 0, 0},
        {1, 1, 1},
        {0, 0, 0}
    },
    {
        {0, 0, 1},
        {1, 1, 1},
        {0, 0, 0}
    }
};

int colors[7] = {1, 2, 3, 4, 5, 6, 7};
int board[BOARD_HEIGHT][BOARD_WIDTH] = {0};
int currentPiece[PIECE_SIZE][PIECE_SIZE];
int currentColor;
int currentX, currentY;
int level = 1;
int linesCleared = 0;
int dropSpeed = 1000; 
int score = 0;

void clearScreen() {
    system("cls");
}

void setColor(int color) {
    SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), color);
}

void spawnPiece() {
    for (int y = 0; y < PIECE_SIZE; y++) {
        for (int x = 0; x < PIECE_SIZE; x++) {
            currentPiece[y][x] = nextPiece[y][x];
        }
    }
    currentColor = nextColor;

    //Generate a new next piece.
    int type = rand() % 7;
    for (int y = 0; y < PIECE_SIZE; y++) {
        for (int x = 0; x < PIECE_SIZE; x++) {
            nextPiece[y][x] = tetrominoes[type][y][x];
        }
    }
    nextColor = colors[type];

    currentX = BOARD_WIDTH / 2 - 2;
    currentY = 0;

    if (checkCollision(currentX, currentY)) {
        gameOver();
    /*if (checkCollision(currentX, currentY)) {
        printf("Game Over\n");
        exit(0); */
    }
}

//Function to display the next piece preview in the top-right corner.
void drawNextPiece() {
    int offsetX = BOARD_WIDTH * 2 + 4;
    int offsetY = 2;

    for (int i = 0; i < offsetY; i++) {
        printf("\n");
    }
    for (int i = 0; i < offsetX; i++) {
        printf(" ");
    }
    printf("Next Piece:\n");

    for (int y = 0; y < PIECE_SIZE; y++) {
        for (int i = 0; i < offsetX; i++) {
            printf(" ");
        }
        for (int x = 0; x < PIECE_SIZE; x++) {
            if (nextPiece[y][x]) {
                setColor(nextColor);
                printf("[]");
            } else {
                printf("  ");
            }
        }
        printf("\n"); 
    }
    setColor(7);
}

//Draw the game board with the updated display area and score.
void drawBoard() {
    clearScreen(); 
    printf(" Tetris-C                 Level: %d\n", level); 
    //Top border
    for (int i = 0; i < DISPLAY_WIDTH; i++) {
        printf("*");
    }
    printf("\n");

    for (int y = 0; y < BOARD_HEIGHT; y++) {
        printf("* "); //Left border
        for (int x = 0; x < BOARD_WIDTH; x++) {
            int cell = board[y][x];

            if (y >= currentY && y < currentY + PIECE_SIZE && x >= currentX && x < currentX + PIECE_SIZE) {
                int pieceY = y - currentY;
                int pieceX = x - currentX;
                if (currentPiece[pieceY][pieceX]) {
                    cell = currentColor;
                }
            }

            if (cell) {
                setColor(cell);
                printf("[]");
            } else {
                printf("  ");
            }
        }
        //Print spaces for alignment and the "Next Piece" and score area
        printf(" *    "); 
        if (y == 1) {
            printf("Next Piece:");
        } else if (y >= 3 && y < 3 + PIECE_SIZE) {
            int nextY = y - 3;
            for (int x = 0; x < PIECE_SIZE; x++) {
                if (nextPiece[nextY][x]) {
                    setColor(nextColor);
                    printf("[]");
                } else {
                    printf("  ");
                }
            }
        } else if (y == 8) { //Score display row
            printf("Score: %d", score);
        }
        printf("\n");
    }

    //Bottom border
    for (int i = 0; i < DISPLAY_WIDTH; i++) {
        printf("*");
    }
    printf("\n");

    setColor(7);
}

void updateScore(int linesClearedInOneMove) {
    switch (linesClearedInOneMove) {
        case 1:
            score += 40 * level;
            break;
        case 2:
            score += 100 * level;
            break;
        case 3:
            score += 300 * level;
            break;
        case 4:
            score += 1200 * level;
            break;
        default:
            break;
    }
}

//Check for and remove fully occupied lines from the board, shifting lines down.
void removeFullLines() {
    int linesClearedInThisMove = 0;

    for (int y = 0; y < BOARD_HEIGHT; y++) {
        int full = 1;
        for (int x = 0; x < BOARD_WIDTH; x++) {
            if (!board[y][x]) {
                full = 0; 
                break;
            }
        }
        if (full) { 
            linesCleared++;
            linesClearedInThisMove++;

            //Shift all rows above the current one down by one.
            for (int yy = y; yy > 0; yy--) {
                for (int x = 0; x < BOARD_WIDTH; x++) {
                    board[yy][x] = board[yy - 1][x];
                }
            }
            //Clear the top row after shifting.
            for (int x = 0; x < BOARD_WIDTH; x++) {
                board[0][x] = 0;
            }
        }
    }
    updateScore(linesClearedInThisMove);

    if (linesCleared >= level * 5) {
        level++;
        dropSpeed -= 50;
        if (dropSpeed < 100) {
            dropSpeed = 100; 
        }
    }
}

void placePiece() {
    for (int y = 0; y < PIECE_SIZE; y++) {
        for (int x = 0; x < PIECE_SIZE; x++) {
            if (currentPiece[y][x]) {
                board[currentY + y][currentX + x] = currentColor;
            }
        }
    }
    score += 5;
}

int checkCollision(int newX, int newY) {
    for (int y = 0; y < PIECE_SIZE; y++) {
        for (int x = 0; x < PIECE_SIZE; x++) {
            if (currentPiece[y][x]) { 
                int boardX = newX + x;
                int boardY = newY + y;

                if (boardX < 0 || boardX >= BOARD_WIDTH || boardY >= BOARD_HEIGHT) {
                    return 1;
                }

                if (boardY >= 0 && board[boardY][boardX]) {
                    return 1;
                }
            }
        }
    }
    return 0;
}

//Rotates the current piece 90 degrees clockwise.
void rotatePiece() {
    int temp[PIECE_SIZE][PIECE_SIZE] = {0};
    for (int y = 0; y < PIECE_SIZE; y++) {
        for (int x = 0; x < PIECE_SIZE; x++) {
            temp[x][PIECE_SIZE - 1 - y] = currentPiece[y][x];
        }
    }

    //Check if the rotation causes a collision.
    if (!checkCollision(currentX, currentY)) {
        for (int y = 0; y < PIECE_SIZE; y++) {
            for (int x = 0; x < PIECE_SIZE; x++) {
                currentPiece[y][x] = temp[y][x];
            }
        }
        while (currentX < 0) currentX++; 
        while (currentX + PIECE_SIZE > BOARD_WIDTH) currentX--;
    }
}

void update() {
    if (!checkCollision(currentX, currentY + 1)) {
        currentY++;
    } else {
        placePiece();
        removeFullLines();
        spawnPiece();
    }
}

void handleInput() {
    if (_kbhit()) {
        char key = _getch();
        switch (key) {
            case 'a':
                if (!checkCollision(currentX - 1, currentY)) {
                    currentX--;
                }
                break;
            case 'd':
                if (!checkCollision(currentX + 1, currentY)) {
                    currentX++;
                }
                break;
            case 's':
                update();
                break;
            case 'w':
                rotatePiece();
                break;
        }
    }
}

void resetGame() {
    for (int y = 0; y < BOARD_HEIGHT; y++) {
        for (int x = 0; x < BOARD_WIDTH; x++) {
            board[y][x] = 0;
        }
    }
    score = 0;
    level = 1;
    linesCleared = 0;
    dropSpeed = 1000;
    spawnPiece();
}

void gameOver() {
    clearScreen();
    printf("Game Over!\n");
    printf("Final Score: %d\n", score);
    printf("Would you like to play again? (Y/N): ");
    
    char choice;
    while (1) {
        choice = _getch();
        if (choice == 'Y' || choice == 'y') {
            resetGame();
            return;
        } else if (choice == 'N' || choice == 'n') {
            printf("\nThank you for playing!\n");
            exit(0);
        }
    }
}

int main() {
    srand(time(NULL));

    int type = rand() % 7;
    for (int y = 0; y < PIECE_SIZE; y++) {
        for (int x = 0; x < PIECE_SIZE; x++) {
            nextPiece[y][x] = tetrominoes[type][y][x];
        }
    }
    nextColor = colors[type];
    spawnPiece();

    unsigned long lastUpdate = GetTickCount();
    while (1) {
        drawBoard();
        handleInput();

        unsigned long now = GetTickCount();
        if (now - lastUpdate >= dropSpeed) {
            update();
            lastUpdate = now;
        }
        Sleep(50);
    }
    return 0;
}
